/* bmdrv2blccc.c
 * BLINKENmini converter for blccc
 *  - gets key-events from bmdrv and sends udp packets with isdn-state to blccc
 * Copyright (C) 2002 sphaera & 1stein (http://blinkenmini.1stein.no-ip.com/)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>

#define VerMaj 1
#define VerMin 0
#define VerRev 0

#define WaitTime 20000
#define IsdnStateSendTime 20

#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))

//keys of controllers
#define CtrCnt 2
#define KeyCnt 12
#define KeyB 0
#define KeyY 1
#define KeySelect 2
#define KeyStart 3
#define KeyUp 4
#define KeyDown 5
#define KeyLeft 6
#define KeyRight 7
#define KeyA 8
#define KeyX 9
#define KeyL 10
#define KeyR 11

//names of buttons
char * ButtonNames[] = { "B", "Y", "Select", "Start", "Up", "Down", "Left", "Right", "A", "X", "L", "R" };

//key-value to isdn-state translation table
struct tTransTab
{
	unsigned long KeyVal;
	unsigned char IsdnState;
} TransTab[] =
{
	{ 0x050, '1' }, //up, left
	{ 0x010, '2' }, //up
	{ 0x090, '3' }, //up, right
	{ 0x040, '4' }, //left
	{ 0x002, '5' }, //Y
	{ 0x080, '6' }, //right
	{ 0x060, '7' }, //down, left
	{ 0x020, '8' }, //down
	{ 0x0A0, '9' }, //down, right
	{ 0x041, '*' }, //B, left
	{ 0x021, '0' }, //B, down
	{ 0x081, '#' }, //B, right
};

//current key-state
int KeyState[CtrCnt][KeyCnt];

//current offhook-state
int Offhook[CtrCnt];

//current isdn-state
unsigned char IsdnState[CtrCnt];

//print that a key was pressed or released
void PrintKeyEvent( unsigned char Ctr, unsigned char Key, unsigned char Pressed )
{
	//display message
	if( Key < count( ButtonNames ) )
		if( Pressed )
			printf( "bmdrv2blccc: button %s of controller %d was pressed\n", ButtonNames[Key], Ctr );
		else
			printf( "bmdrv2blccc: button %s of controller %d was released\n", ButtonNames[Key], Ctr );
	else
		if( Pressed )
			printf( "bmdrv2blccc: button %d of controller %d was pressed\n", Key, Ctr );
		else
			printf( "bmdrv2blccc: button %d of controller %d was released\n", Key, Ctr );
}

//print key state
void PrintKeyState( )
{
	unsigned char Ctr, Key;

	printf( "bmdrv2blccc: new key state =" );
	for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
	{
		printf( " " );
		for( Key = 0; Key < KeyCnt; Key++ )
		{
			if( KeyState[Ctr][Key] )
				printf( "*" );
			else
				printf( "." );
		}
	}
	printf( "\n" );
};

//print offhook state
void PrintOffhook( )
{
	unsigned char Ctr;

	printf( "bmdrv2blccc: new offhook state =" );
	for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
		if( Offhook[Ctr] )
			printf( " offhook" );
		else
			printf( " onhook" );
	printf( "\n" );
}

//print isdn state
void PrintIsdnState( )
{
	unsigned char Ctr;

	printf( "bmdrv2blccc: new isdn state =" );
	for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
		if( IsdnState[Ctr] >= ' ' && IsdnState[Ctr] <= '~' )
			printf( " 0x%02X (%c)", IsdnState[Ctr], IsdnState[Ctr] );
		else
			printf( " 0x%02X    ", IsdnState[Ctr] );
	printf( "\n" );
}

//main program
int main( int ArgC, char * * Args )
{
	//local variables
	int BmDrv, Socket, Ip, Port, I, KeyStateChange, OffhookChange, IsdnStateChange, IsdnStateSendCnt;
	char * pChr;
	unsigned char Ctr, Key, Pressed, NewIsdnState;
	struct sockaddr_in Addr;
	unsigned char Buffer[8192], BufferLen;
	unsigned long KeyVal;
	struct hostent * pHostent;

	//print message
	printf( "\n" );
	printf( "bmdrv2blccc: BLINKENmini converter for blccc (version %d.%d.%d)\n", VerMaj, VerMin, VerRev );
	printf( "             gets key-events from bmdrv and sends udp packets with isdn-state to blccc\n" );
	printf( "             done in 2002 by sphaera & 1stein\n" );
	printf( "             http://blinkenmini.1stein.no-ip.com/\n" );
	printf( "             distributed under GNU Public License - no warranty\n\n" );

	//check parameters
	if( ArgC != 3 )
	{
		//diplay help message
		printf( "syntax: bmdrv2blccc <[host:]port> <in-bmdrv>\n" );
		printf( "  - [host:]port - host and port to send data to\n" );
		printf( "  - in-bmdrv - path to input device of bmdrv\n\n" );
		return -1;
	}

	//parse ip and port
	pChr = strchr( Args[1], ':' );
	if( pChr == NULL )
	{
		((unsigned char *)&Ip)[0] = 127; //use localhost
		((unsigned char *)&Ip)[1] = 0;
		((unsigned char *)&Ip)[2] = 0;
		((unsigned char *)&Ip)[3] = 1;
		Port = atoi( Args[1] );
	}
	else
	{
		*pChr = 0;
		Ip = inet_addr( Args[1] );
		Port = atoi( pChr + 1 );
		//resolve hostname
		if( Ip == INADDR_NONE )
		{
			pHostent = gethostbyname( Args[1] );
			if( pHostent != NULL )
				if( pHostent->h_addr_list != NULL )
					if( pHostent->h_addr_list[0] != NULL )
						Ip = *(unsigned long *)(pHostent->h_addr_list[0]);
		}
	}	
	//no ip found
	if( Ip == INADDR_NONE )
	{
		printf( "\nbmdrv2blccc: could not resolve host\n\n" );
		return -1;
	}

	//open bmdrv
	BmDrv = open( Args[2], O_RDONLY | O_NONBLOCK );
	if( BmDrv == -1 )
	{
		printf( "\nbmdrv2blccc: could not open in-bmdrv: %s\n\n", strerror( errno ) );
		return -1;
	}


	//create socket
	Socket = socket( AF_INET, SOCK_DGRAM, 0 );
	if( Socket == -1 )
	{
		printf( "\nbmdrv2blccc: could not create socket: %s\n\n", strerror( errno ) );
		close( BmDrv );
		return -1;
	}

	//reset key-state
	for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
		for( Key = 0; Key < KeyCnt; Key++ )
			KeyState[Ctr][Key] = 0;
	//reset offhook-state
	for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
		Offhook[Ctr] = 0;
	//reset isdn-state
	for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
		IsdnState[Ctr] = 0x00;

	//main loop
	BufferLen = 0;
	IsdnStateSendCnt = 0;
	for( ; ; )
	{
		//read data from bmdrv
		I = read( BmDrv, Buffer + BufferLen, sizeof( Buffer ) - BufferLen );
		if( I == -1 && errno != EAGAIN )
		{
			printf( "\nbmdrv2blccc: error while reading from in-bmdrv: %s\n\n", strerror( errno ) );
			close( Socket );
			close( BmDrv );
			return -1;
		}
		if( I > 0 )
			BufferLen += I;

		//process data from bmdrv
		KeyStateChange = 0;
		for( I = 0; I < BufferLen; )
		{
			//find next 0xFF
			for( ; I < BufferLen && Buffer[I] != 0xFF; I++ );
			//3 bytes behind 0xFF
			if( I + 3 < BufferLen )
			{
				//get data
				Pressed = Buffer[I + 1];
				Ctr = Buffer[I + 2];
				Key = Buffer[I + 3];
				//print key event
				PrintKeyEvent( Ctr, Key, Pressed );
				//check data
				if( Pressed <= 1 && Ctr <= CtrCnt && Key <= KeyCnt )
				{
					//update key state
					KeyState[Ctr][Key] = Pressed;
					//key state changed
					KeyStateChange = 1;
				}
				//4 bytes processed
				I += 4;
			} //if( I + 3 < BufferLen )
		} //for( I ...
		//remove processed data from buffer
		BufferLen -= I;
		memmove( Buffer, Buffer + I, BufferLen );

		//print new key-state if it changed
		if( KeyStateChange )
			//print new key state
			PrintKeyState( KeyState );

		//generate new offhook-states and remeber if any changed
		OffhookChange = 0;
		for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
		{
			//Start pressed and Select released
			if( KeyState[Ctr][KeyStart] && !KeyState[Ctr][KeySelect])
			{
				//offhook
				if( Offhook[Ctr] == 0 )
					OffhookChange = 1;
				Offhook[Ctr] = 1;
			}
			//Select pressed and Start released
			if( KeyState[Ctr][KeySelect] && !KeyState[Ctr][KeyStart] )
			{
				//onhook
				if( Offhook[Ctr] != 0 )
					OffhookChange = 1;
				Offhook[Ctr] = 0;
			}
		}

		//print offhook-state if it changed
		if( OffhookChange )
			PrintOffhook( );

		//generate new isdn-states and remember if any changed
		IsdnStateChange = 0;
		for( Ctr = 0; Ctr < CtrCnt; Ctr++ )
		{
			//put all keys as bits into a single variable
			KeyVal = 0;
			for( Key = KeyCnt - 1; (char)Key >= 0; Key-- )
			{
				KeyVal <<= 1;
				if( KeyState[Ctr][Key] && Key != KeyStart && Key != KeySelect ) //exclude Start and Select
					KeyVal |= 1;
			}				
			//get new isdn-state
			if( Offhook[Ctr] )
			{
				NewIsdnState = 0xFF; //default isdn-state for offhook
				for( I = 0; I < count( TransTab ); I++ )
					if( TransTab[I].KeyVal == KeyVal )
						NewIsdnState = TransTab[I].IsdnState;
			}
			else
				NewIsdnState = 0x00; //isdn-state for onhook
			//isdn-state changed
			if( IsdnState[Ctr] != NewIsdnState )
			{
				IsdnState[Ctr] = NewIsdnState;
				IsdnStateChange = 1;
			}
		}

		//print isdn-state if it changed
		if( IsdnStateChange )
			PrintIsdnState( );

		//isdn-state changed or long time not send
		IsdnStateSendCnt++;
		if( IsdnStateChange || IsdnStateSendCnt >= IsdnStateSendTime )
		{
			IsdnStateSendCnt = 0;
			//send udp packet with isdn-state to address
			Addr.sin_family = AF_INET;
			Addr.sin_port = htons( Port );
			Addr.sin_addr.s_addr = Ip;	
			I = sendto( Socket, IsdnState, CtrCnt, 0, (struct sockaddr *)&Addr, sizeof( struct sockaddr ) );
			if( I == -1 )
			{
				printf( "\nbmdrv2blccc: could not send udp packet: %s\n\n", strerror( errno ) );
				close( Socket );
				close( BmDrv );
				return -1;
			}
		}

		//wait a little bit
		usleep( WaitTime );
	} //for( ; ; )

	//close socket
	close( Socket );

	//close bmdrv
	close( BmDrv );

	//print end-message
	printf( "\nbmdrv2blccc: done...\n\n" );

	return 0;
}
